from typing import cast, Any, Callable, Deque, Dict, Generator, List, Sequence, Set, Tuple, Type, Union

import asyncio
from collections import deque

from hedgehog.protocol.messages import Message, ack, analog, digital, io, motor, process, servo


UpdateHandler = Generator[None, Message, None]
UpdateKey = Tuple[Type[Message], Any]
AsyncHandler = Callable[[Message], Tuple[Set[UpdateKey], UpdateHandler]]


class ProcessHandler:
    def __init__(self, on_stdout, on_stderr, on_exit, sequential=True):
        self._on_stdout = on_stdout
        self._on_stderr = on_stderr
        self._on_exit = on_exit
        self._sequential = sequential
        pass

    async def handle_update(self, update: Union[process.StreamUpdate, process.ExitUpdate]) -> None:
        if isinstance(update, process.StreamUpdate):
            if update.fileno == process.STDOUT:
                if self._on_stdout is not None:
                    await self._on_stdout(self.pid, update.fileno, update.chunk)
            else:
                if self._on_stderr is not None:
                    await self._on_stderr(self.pid, update.fileno, update.chunk)
        elif isinstance(update, process.ExitUpdate):
            if self._on_exit is not None:
                await self._on_exit(self.pid, update.exit_code)
        else:  # pragma: nocover
            assert False, update

    def _handle_updates_sequential(self, tasks: List[asyncio.Task]):
        queue = asyncio.Queue()

        async def run_updates() -> None:
            while True:
                update = await queue.get()
                await self.handle_update(update)
                if isinstance(update, process.ExitUpdate):
                    break

        tasks.append(asyncio.create_task(run_updates()))

        while True:
            update = yield
            queue.put_nowait(update)
            if isinstance(update, process.ExitUpdate):
                break

    def _handle_updates_concurrent(self, tasks: List[asyncio.Task]):
        while True:
            update = yield
            tasks.append(asyncio.create_task(self.handle_update(update)))
            if isinstance(update, process.ExitUpdate):
                break

    def _handle_updates(self):
        tasks: List[asyncio.Task] = []

        try:
            if self._sequential:
                yield from self._handle_updates_sequential(tasks)
            else:
                yield from self._handle_updates_concurrent(tasks)

            # here we expect shutdown
            try:
                yield
            except StopIteration:
                raise
            else:
                raise RuntimeError("generator didn't stop")
        finally:
            for task in tasks:
                task.cancel()

    def __call__(self, reply: process.ExecuteReply) -> Tuple[Set[UpdateKey], UpdateHandler]:
        self.pid = reply.pid
        update_keys = {(process.StreamUpdate, self.pid), (process.ExitUpdate, self.pid)}
        update_handler: UpdateHandler = self._handle_updates()
        next(update_handler)
        return update_keys, update_handler


class HandlerRegistry(object):
    _update_keys = {
        io.CommandUpdate: lambda update: cast(io.CommandUpdate, update).port,
        analog.Update: lambda update: cast(analog.Update, update).port,
        digital.Update: lambda update: cast(digital.Update, update).port,
        motor.CommandUpdate: lambda update: cast(motor.CommandUpdate, update).port,
        motor.StateUpdate: lambda update: cast(motor.StateUpdate, update).port,
        servo.CommandUpdate: lambda update: cast(servo.CommandUpdate, update).port,
        process.StreamUpdate: lambda update: cast(process.StreamUpdate, update).pid,
        process.ExitUpdate: lambda update: cast(process.ExitUpdate, update).pid,
    }  # type: Dict[Type[Message], Callable[[Message], Any]]

    @staticmethod
    def _update_key(update: Message) -> UpdateKey:
        cls = type(update)
        return cls, HandlerRegistry._update_keys[cls](update)

    def __init__(self) -> None:
        self._queue: Deque[Sequence[AsyncHandler]] = deque()
        self._handlers = {}  # type: Dict[UpdateKey, UpdateHandler]

    def prepare_register(self, handlers: Sequence[AsyncHandler]):
        self._queue.append(handlers)

    def complete_register(self, replies: Sequence[Message]):
        handlers = self._queue.popleft()
        assert len(handlers) == len(replies)
        for handler, reply in zip(handlers, replies):
            if handler is None:
                continue
            if isinstance(reply, ack.Acknowledgement) and reply.code != ack.OK:
                # we simply don't call the EventHandler and everything is fine
                continue

            update_keys, update_handler = handler(reply)
            for update_key in update_keys:
                if update_key in self._handlers:
                    self._handlers[update_key].close()
                self._handlers[update_key] = update_handler

    def handle_updates(self, updates: Sequence[Message]) -> None:
        for update in updates:
            update_key = HandlerRegistry._update_key(update)
            if update_key in self._handlers:
                self._handlers[update_key].send(update)

    def shutdown(self) -> None:
        for handler in self._handlers.values():
            handler.close()
        self._handlers.clear()



from typing import Awaitable, Callable, Optional, Tuple, TypeVar

import concurrent.futures
import logging
import os
import signal
import threading
import time
import zmq.asyncio
from contextlib import contextmanager, ExitStack
from functools import partial

from concurrent_utils.event_loop_thread import EventLoopThread
from hedgehog.protocol.messages import motor
from . import async_client, shutdown_handler

logger = logging.getLogger(__name__)

T = TypeVar('T')


class SyncClient(object):
    def __init__(self, ctx: zmq.asyncio.Context, endpoint: str='tcp://127.0.0.1:10789') -> None:
        self._loop = EventLoopThread()
        self.ctx = ctx
        self.endpoint = endpoint
        self.client = None  # type: async_client.AsyncClient
        self._signal_ctx = None

    def _create_client(self):
        return async_client.AsyncClient(self.ctx, self.endpoint)  # pragma: nocover

    def _call(self, coro: Awaitable[T]) -> T:
        return self._loop.run_coroutine(coro).result()

    def _call_safe(self, coro_fun: Callable[[], Awaitable[T]]) -> T:
        if self.client is None or self.client.is_closed:
            raise RuntimeError("The client is not active, use `with client:`")
        return self._call(coro_fun())

    def _enter(self, daemon=False):
        if self.client and self.client.is_shutdown:
            # necessary because we can't use the event loop thread - which is already closed in this case -
            # to try and start the client, which would then fail with this error
            raise RuntimeError("Cannot reuse a client after it was once shut down")
        elif self.client:
            self._call(self.client._aenter(daemon=daemon))
            return self
        else:
            with ExitStack() as enter_stack:
                enter_stack.enter_context(self._loop)

                async def create_client():
                    return self._create_client()

                # create the client on the event loop, to be sure the client uses the correct one
                self.client = self._call(create_client())

                @enter_stack.callback
                def clear_client():
                    self.client = None

                if threading.current_thread() is threading.main_thread():
                    def sigint_handler(signal, frame):
                        self.shutdown()

                    signal_ctx = shutdown_handler.register(signal.SIGINT, sigint_handler)
                    enter_stack.enter_context(signal_ctx)
                else:
                    signal_ctx = None

                self._call(self.client._aenter(daemon=daemon))
                self._signal_ctx = signal_ctx

                # all went well, so don't clean up everything
                enter_stack.pop_all()
                return self

    def _exit(self, exc_type, exc_val, exc_tb, daemon=False):
        stack = ExitStack()

        # called last
        @stack.push
        def exit_loop(exc_type, exc_val, exc_tb):
            if self.client.is_closed:
                return self._loop.__exit__(exc_type, exc_val, exc_tb)

        if threading.current_thread() is threading.main_thread():
            # TODO the main thread is not necessarily the last thread to finish.
            # Should the signal handler be removed in case it isn't?
            stack.push(self._signal_ctx)

        # called first
        # exit the client with the given daemon-ness, maybe leading the client to close
        @stack.push
        def exit_client(exc_type, exc_val, exc_tb):
            return self._call(self.client._aexit(exc_type, exc_val, exc_tb, daemon=daemon))

        return stack.__exit__(exc_type, exc_val, exc_tb)

    def __enter__(self):
        return self._enter()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._exit(exc_type, exc_val, exc_tb)

    @property
    @contextmanager
    def daemon(self):
        with ExitStack() as stack:
            ret = self._enter(daemon=True)
            stack.push(partial(self._exit, daemon=True))
            yield ret

    @property
    def is_shutdown(self):
        return self.client is not None and self.client.is_shutdown

    @property
    def is_closed(self):
        return self.client is not None and self.client.is_closed

    def shutdown(self) -> None:
        self._call_safe(lambda: self.client.shutdown())

    def spawn(self, callback, *args, name=None, daemon=False, **kwargs) -> threading.Thread:
        future = concurrent.futures.Future()

        def target(*args, **kwargs):
            with (self.daemon if daemon else self):
                future.set_result(None)
                callback(*args, **kwargs)

        result = threading.Thread(target=target, name=name, args=args, kwargs=kwargs)
        result.start()
        future.result()
        return result


class HedgehogClientMixin(object):
    def _create_client(self):
        return async_client.HedgehogClient(self.ctx, self.endpoint)

    def set_input_state(self, port: int, pullup: bool) -> None:
        self._call_safe(lambda: self.client.set_input_state(port, pullup))

    def get_analog(self, port: int) -> int:
        return self._call_safe(lambda: self.client.get_analog(port))

    def get_digital(self, port: int) -> bool:
        return self._call_safe(lambda: self.client.get_digital(port))

    def set_digital_output(self, port: int, level: bool) -> None:
        self._call_safe(lambda: self.client.set_digital_output(port, level))

    def get_io_config(self, port: int) -> int:
        return self._call_safe(lambda: self.client.get_io_config(port))

    def configure_motor(self, port: int, config: motor.Config) -> int:
        self._call_safe(lambda: self.client.configure_motor(port, config))

    def configure_motor_dc(self, port: int) -> int:
        self._call_safe(lambda: self.client.configure_motor_dc(port))

    def configure_motor_encoder(self, port: int, encoder_a_port: int, encoder_b_port: int) -> int:
        self._call_safe(lambda: self.client.configure_motor_encoder(port, encoder_a_port, encoder_b_port))

    def configure_motor_stepper(self, port: int) -> int:
        self._call_safe(lambda: self.client.configure_motor_stepper(port))

    def set_motor(self, port: int, state: int, amount: int=0,
                  reached_state: int=motor.POWER, relative: int=None, absolute: int=None,
                  on_reached: Callable[[int, int], None]=None) -> None:
        self._call_safe(lambda: self.client.set_motor(port, state, amount, reached_state, relative, absolute, on_reached))

    def move(self, port: int, amount: int, state: int=motor.POWER) -> None:
        self._call_safe(lambda: self.client.move(port, amount, state))

    def move_relative_position(self, port: int, amount: int, relative: int, state: int=motor.POWER,
                               on_reached: Callable[[int, int], None]=None) -> None:
        self._call_safe(lambda: self.client.move_relative_position(port, amount, relative, state, on_reached))

    def move_absolute_position(self, port: int, amount: int, absolute: int, state: int=motor.POWER,
                               on_reached: Callable[[int, int], None]=None) -> None:
        self._call_safe(lambda: self.client.move_absolute_position(port, amount, absolute, state, on_reached))

    def get_motor_command(self, port: int) -> Tuple[int, int]:
        return self._call_safe(lambda: self.client.get_motor_command(port))

    def get_motor_state(self, port: int) -> Tuple[int, int]:
        return self._call_safe(lambda: self.client.get_motor_state(port))

    def get_motor_velocity(self, port: int) -> int:
        return self._call_safe(lambda: self.client.get_motor_velocity(port))

    def get_motor_position(self, port: int) -> int:
        return self._call_safe(lambda: self.client.get_motor_position(port))

    def set_motor_position(self, port: int, position: int) -> None:
        self._call_safe(lambda: self.client.set_motor_position(port, position))

    def set_servo(self, port: int, active: bool, position: int) -> None:
        self._call_safe(lambda: self.client.set_servo(port, active, position))

    def get_servo_command(self, port: int) -> Tuple[bool, int]:
        return self._call_safe(lambda: self.client.get_servo_command(port))

    def get_imu_rate(self) -> Tuple[int, int, int]:
        return self._call_safe(lambda: self.client.get_imu_rate())

    def get_imu_acceleration(self) -> Tuple[int, int, int]:
        return self._call_safe(lambda: self.client.get_imu_acceleration())

    def get_imu_pose(self) -> Tuple[int, int, int]:
        return self._call_safe(lambda: self.client.get_imu_pose())

    def execute_process(self, *args: str, working_dir: str=None, on_stdout=None, on_stderr=None, on_exit=None) -> int:
        return self._call_safe(
            lambda: self.client.execute_process(*args, working_dir=working_dir,
                                                on_stdout=on_stdout, on_stderr=on_stderr, on_exit=on_exit))

    def signal_process(self, pid: int, signal: int=2) -> None:
        self._call_safe(lambda: self.client.signal_process(pid, signal))

    def send_process_data(self, pid: int, chunk: bytes=b'') -> None:
        self._call_safe(lambda: self.client.send_process_data(pid, chunk))

    def set_speaker(self, frequency: Optional[int]) -> None:
        self._call_safe(lambda: self.client.set_speaker(frequency))


class HedgehogClient(HedgehogClientMixin, SyncClient):
    pass


@contextmanager
def connect(endpoint='tcp://127.0.0.1:10789', emergency=None,
            ctx=None, client_class=HedgehogClient, process_setup=True):
    # TODO SIGINT handling

    ctx = ctx or zmq.asyncio.Context()
    with client_class(ctx, endpoint) as client:
        # TODO a remote application's emergency_stop is remote, so it won't work in case of a disconnection!
        def emergency_stop():
            client.set_input_state(emergency, True)
            # while not client.get_digital(emergency):
            while client.get_digital(emergency):
                time.sleep(0.1)

            os.kill(os.getpid(), signal.SIGINT)

        if emergency is not None:
            client.spawn(emergency_stop, name="emergency_stop", daemon=True)

        yield client



from typing import cast, Any, Awaitable, Callable, Deque, List, Optional, Sequence, Tuple

import asyncio
import logging
import os
import signal
import threading
import zmq.asyncio
from collections import deque
from contextlib import asynccontextmanager, AsyncExitStack
from functools import partial

from concurrent_utils.pipe import PipeEnd
from concurrent_utils.component import Component, component_coro_wrapper, start_component
from hedgehog.protocol import errors, ClientSide
from hedgehog.protocol.zmq.asyncio import DealerRouterSocket
from hedgehog.protocol.messages import Message, ack, io, analog, digital, motor, servo, imu, process, speaker
from . import shutdown_handler
from .async_handlers import AsyncHandler, HandlerRegistry, ProcessHandler

logger = logging.getLogger(__name__)


class AsyncClient:
    def __init__(self, ctx: zmq.asyncio.Context, endpoint: str='tcp://127.0.0.1:10789') -> None:
        self.ctx = ctx
        self.endpoint = endpoint
        self.registry = HandlerRegistry()
        self.socket = None  # type: DealerRouterSocket
        self._reply_condition = asyncio.Condition()
        self._replies: Deque[Sequence[Message]] = deque()

        self._open_count = 0
        self._daemon_count = 0
        self._exit_stack = None  # type: AsyncExitStack
        self._shutdown = False

    async def _aenter(self, daemon=False):
        async with AsyncExitStack() as enter_stack:
            if daemon and self._open_count == 0:
                raise RuntimeError("The client is not active, first use of the client must not be daemon")
            if self._shutdown:
                raise RuntimeError("Cannot reuse a client after it was once shut down")

            logger.debug("Entering as %s", "daemon" if daemon else "regular")

            self._open_count += 1

            @enter_stack.callback
            def decrement_open_count():
                self._open_count -= 1

            if daemon:
                self._daemon_count += 1

                # testing: I see no good way to cause a fault that triggers this...
                # the code is almost the same as decrement_open_count, so ignore it for coverage
                @enter_stack.callback
                def decrement_daemon_count():
                    self._daemon_count -= 1  # pragma: nocover

            if self._open_count == 1:
                logger.debug("Activating client...")

                async with AsyncExitStack() as stack:
                    @asynccontextmanager
                    async def start() -> Component[None]:
                        component = await start_component(self.workload)
                        try:
                            yield component
                        finally:
                            await component.stop()
                    await stack.enter_async_context(start())

                    if threading.current_thread() is threading.main_thread():
                        loop = asyncio.get_event_loop()

                        def sigint_handler():
                            task = loop.create_task(self.shutdown())

                            # if this signal handler is called during _aenter, register the await with `stack`;
                            # otherwise, with `self._exit_stack`
                            exit_stack = self._exit_stack if self._exit_stack is not None else stack

                            @exit_stack.push_async_callback
                            async def await_shutdown():
                                await task

                        stack.enter_context(shutdown_handler.register_async(signal.SIGINT, sigint_handler))

                    # save the exit actions that need undoing...
                    self._exit_stack = stack.pop_all()

            # ...and discard those that were only for the error case
            enter_stack.pop_all()
            logger.debug("Open: %d (%d daemon)", self._open_count, self._daemon_count)
            return self

    async def _aexit(self, exc_type, exc_val, exc_tb, daemon=False):
        stack = AsyncExitStack()

        # called last
        @stack.callback
        def decrement_open_count():
            self._open_count -= 1
            logger.debug("Open: %d (%d daemon)", self._open_count, self._daemon_count)

        @stack.push_async_exit
        async def exit(exc_type, exc_val, exc_tb):
            if self._open_count == 1:
                logger.debug("Deactivating client...")
                try:
                    return await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)
                finally:
                    self._exit_stack = None

        @stack.push_async_callback
        async def shutdown():
            if self._open_count - 1 == self._daemon_count:
                await self.shutdown()

        @stack.callback
        def decrement_daemon_count():
            if daemon:
                self._daemon_count -= 1

        # called first
        @stack.push
        def suppress_shutdown_error(exc_type, exc_val, exc_tb):
            if exc_type == errors.EmergencyShutdown:
                print(exc_val)
                return True

        logger.debug("Exiting as %s", "daemon" if daemon else "regular")

        return await stack.__aexit__(exc_type, exc_val, exc_tb)

    async def __aenter__(self):
        return await self._aenter()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self._aexit(exc_type, exc_val, exc_tb)

    @property
    @asynccontextmanager
    async def daemon(self):
        async with AsyncExitStack() as stack:
            ret = await self._aenter(daemon=True)
            stack.push_async_exit(partial(self._aexit, daemon=True))
            yield ret

    @property
    def is_shutdown(self):
        return self._shutdown

    @property
    def is_closed(self):
        return self.is_shutdown and self._open_count == 0

    @property
    @asynccontextmanager
    async def _job(self):
        if self._open_count == 0:
            raise RuntimeError("The client is not active, use `async with client:`")

        async with self._reply_condition:
            yield

    async def _handle_updates(self):
        while True:
            _, msgs = await self.socket.recv_msgs()
            assert len(msgs) > 0
            async with self._reply_condition:
                # either, all messages are replies corresponding to the previous requests,
                # or all messages are asynchronous updates
                if msgs[0].is_async:
                    assert all(msg.is_async for msg in msgs)

                    # handle asynchronous messages
                    logger.debug("Receive updates: %s", msgs)
                    self.registry.handle_updates(msgs)
                else:
                    assert not any(msg.is_async for msg in msgs)

                    # handle synchronous messages
                    self.registry.complete_register(msgs)
                    self._replies.append(msgs)
                    self._reply_condition.notify()

    async def _workload(self, *, commands: PipeEnd, events: PipeEnd) -> None:
        with DealerRouterSocket(self.ctx, zmq.DEALER, side=ClientSide) as self.socket:
            self.socket.connect(self.endpoint)
            await events.send(Component.EVENT_START)

            updates = asyncio.create_task(self._handle_updates())
            try:
                while True:
                    command = await commands.recv()
                    if command == Component.COMMAND_STOP:
                        break
                    else:
                        raise ValueError(f"unknown command: {command!r}")
            finally:
                updates.cancel()
                try:
                    await updates
                except asyncio.CancelledError:
                    pass

    async def workload(self, *, commands: PipeEnd, events: PipeEnd) -> None:
        return await component_coro_wrapper(self._workload, commands=commands, events=events)

    async def send(self, msg: Message, handler: AsyncHandler=None) -> Optional[Message]:
        reply, = await self.send_multipart((msg, handler))
        if isinstance(reply, ack.Acknowledgement):
            if reply.code != ack.OK:
                raise errors.error(reply.code, reply.message)
            return None
        else:
            return reply

    async def send_multipart(self, *cmds: Tuple[Message, AsyncHandler]) -> Any:
        async with self._job:
            if self._shutdown:
                replies = [ack.Acknowledgement(ack.FAILED_COMMAND, "Emergency Shutdown activated") for _ in cmds]
                for _, handler in cmds:
                    if handler is not None:
                        handler.close()
                return replies
            else:
                return await self._send(tuple(request for request, _ in cmds), tuple(handler for _, handler in cmds))

    async def shutdown(self) -> None:
        async with self._job:
            if not self._shutdown:
                self._shutdown = True
                self.registry.shutdown()

                msgs = []  # type: List[Message]
                msgs.extend(motor.Action(port, motor.POWER, 0) for port in range(0, 4))
                msgs.extend(servo.Action(port, False, 0) for port in range(0, 4))
                await self._send(msgs, tuple(None for _ in msgs))

    async def _send(self, requests: Sequence[Message], handlers: Sequence[AsyncHandler]) -> Sequence[Message]:
        logger.debug("Send commands:   %s", requests)
        self.registry.prepare_register(handlers)
        await self.socket.send_msgs((), requests)

        await self._reply_condition.wait()
        replies = self._replies.popleft()
        logger.debug("Receive replies: %s", replies)
        return replies

    async def spawn(self, awaitable: Awaitable[Any], daemon: bool=False) -> asyncio.Task:
        event = asyncio.Event()

        async def task():
            async with (self.daemon if daemon else self):
                event.set()
                await awaitable

        result = asyncio.create_task(task())
        await event.wait()
        return result


class HedgehogClientMixin:
    async def set_input_state(self, port: int, pullup: bool) -> None:
        await self.send(io.Action(port, io.INPUT_PULLUP if pullup else io.INPUT_FLOATING))

    async def get_analog(self, port: int) -> int:
        response = cast(analog.Reply, await self.send(analog.Request(port)))
        assert response.port == port
        return response.value

    async def get_digital(self, port: int) -> bool:
        response = cast(digital.Reply, await self.send(digital.Request(port)))
        assert response.port == port
        return response.value

    async def set_digital_output(self, port: int, level: bool) -> None:
        await self.send(io.Action(port, io.OUTPUT_ON if level else io.OUTPUT_OFF))

    async def get_io_config(self, port: int) -> int:
        response = cast(io.CommandReply, await self.send(io.CommandRequest(port)))
        assert response.port == port
        return response.flags

    async def configure_motor(self, port: int, config: motor.Config) -> int:
        await self.send(motor.ConfigAction(port, config))

    async def configure_motor_dc(self, port: int) -> int:
        await self.configure_motor(port, motor.DcConfig())

    async def configure_motor_encoder(self, port: int, encoder_a_port: int, encoder_b_port: int) -> int:
        await self.configure_motor(port, motor.EncoderConfig(encoder_a_port, encoder_b_port))

    async def configure_motor_stepper(self, port: int) -> int:
        await self.configure_motor(port, motor.StepperConfig())

    async def set_motor(self, port: int, state: int, amount: int=0,
                  reached_state: int=motor.POWER, relative: int=None, absolute: int=None,
                  on_reached: Callable[[int, int], None]=None) -> None:
        # if on_reached is not None:
        #     if relative is None and absolute is None:
        #         raise ValueError("callback given, but no end position")
        #     handler = MotorUpdateHandler(on_reached)
        # else:
        #     handler = None
        await self.send(motor.Action(port, state, amount, reached_state, relative, absolute))

    async def move(self, port: int, amount: int, state: int=motor.POWER) -> None:
        await self.set_motor(port, state, amount)

    async def move_relative_position(self, port: int, amount: int, relative: int, state: int=motor.POWER,
                               on_reached: Callable[[int, int], None]=None) -> None:
        await self.set_motor(port, state, amount, relative=relative, on_reached=on_reached)

    async def move_absolute_position(self, port: int, amount: int, absolute: int, state: int=motor.POWER,
                               on_reached: Callable[[int, int], None]=None) -> None:
        await self.set_motor(port, state, amount, absolute=absolute, on_reached=on_reached)

    async def get_motor_command(self, port: int) -> Tuple[int, int]:
        response = cast(motor.CommandReply, await self.send(motor.CommandRequest(port)))
        assert response.port == port
        return response.state, response.amount

    async def get_motor_state(self, port: int) -> Tuple[int, int]:
        response = cast(motor.StateReply, await self.send(motor.StateRequest(port)))
        assert response.port == port
        return response.velocity, response.position

    async def get_motor_velocity(self, port: int) -> int:
        velocity, _ = await self.get_motor_state(port)
        return velocity

    async def get_motor_position(self, port: int) -> int:
        _, position = await self.get_motor_state(port)
        return position

    async def set_motor_position(self, port: int, position: int) -> None:
        await self.send(motor.SetPositionAction(port, position))

    async def set_servo(self, port: int, active: bool, position: int) -> None:
        await self.send(servo.Action(port, active, position))

    async def get_servo_command(self, port: int) -> Tuple[bool, int]:
        response = cast(servo.CommandReply, await self.send(servo.CommandRequest(port)))
        assert response.port == port
        return response.active, response.position

    async def get_imu_rate(self) -> Tuple[int, int, int]:
        response = cast(imu.RateReply, await self.send(imu.RateRequest()))
        return response.x, response.y, response.z

    async def get_imu_acceleration(self) -> Tuple[int, int, int]:
        response = cast(imu.AccelerationReply, await self.send(imu.AccelerationRequest()))
        return response.x, response.y, response.z

    async def get_imu_pose(self) -> Tuple[int, int, int]:
        response = cast(imu.PoseReply, await self.send(imu.PoseRequest()))
        return response.x, response.y, response.z

    async def execute_process(self, *args: str, working_dir: str=None, on_stdout=None, on_stderr=None, on_exit=None) -> int:
        if on_stdout is not None or on_stderr is not None or on_exit is not None:
            handler = ProcessHandler(on_stdout, on_stderr, on_exit)
        else:
            handler = None
        response = cast(process.ExecuteReply, await self.send(process.ExecuteAction(*args, working_dir=working_dir), handler))
        return response.pid

    async def signal_process(self, pid: int, signal: int=2) -> None:
        await self.send(process.SignalAction(pid, signal))

    async def send_process_data(self, pid: int, chunk: bytes=b'') -> None:
        await self.send(process.StreamAction(pid, process.STDIN, chunk))

    async def set_speaker(self, frequency: Optional[int]) -> None:
        await self.send(speaker.Action(frequency))


class HedgehogClient(HedgehogClientMixin, AsyncClient):
    pass


@asynccontextmanager
async def connect(endpoint='tcp://127.0.0.1:10789', emergency=None,
                  ctx=None, client_class=HedgehogClient, process_setup=True):
    # TODO SIGINT handling

    ctx = ctx or zmq.asyncio.Context()
    async with client_class(ctx, endpoint) as client:
        # TODO a remote application's emergency_stop is remote, so it won't work in case of a disconnection!
        async def emergency_stop():
            await client.set_input_state(emergency, True)
            # while not client.get_digital(emergency):
            while await client.get_digital(emergency):
                await asyncio.sleep(0.1)

            os.kill(os.getpid(), signal.SIGINT)

        if emergency is not None:
            await client.spawn(emergency_stop(), daemon=True)

        yield client



import asyncio
import logging
import signal
import threading
from collections import namedtuple, OrderedDict
from contextlib import contextmanager

logger = logging.getLogger(__name__)


__Handler = namedtuple('__Handler', ('old_handler', 'callbacks'))

__handlers = {}
__async_handlers = {}


@contextmanager
def register(signalnum, callback):
    if threading.current_thread() is not threading.main_thread():
        raise RuntimeError("Signal handlers can only be manipulated from the main thread")

    if signalnum in __async_handlers:
        raise RuntimeError("Can't combine regular and event loop signal handling")

    if signalnum not in __handlers:
        old_handler = signal.getsignal(signalnum)
        if old_handler is None:
            # None means that the previous signal handler was not installed from Python
            # it's not legal to pass None to signal(), so restore the default
            logger.warning("Removing a signal handler that can't be restored")
            if signalnum == signal.SIGINT:
                old_handler = signal.default_int_handler
            else:
                old_handler = signal.SIG_DFL

        handler = __handlers[signalnum] = __Handler(old_handler, OrderedDict())

        if callable(old_handler) and old_handler is not signal.default_int_handler:
            # the old signal handler is a callable; add it to our wrapper handler
            # to retain its behavior.
            handler.callbacks[None] = old_handler

        def do_handle(signalnum, frame):
            for key, callback in reversed(__handlers[signalnum].callbacks.items()):
                try:
                    callback(signalnum, frame)
                except Exception:
                    if key is None:
                        # this is the old signal handler; if we didn't install our own, this would not be caught,
                        # so reraise it here
                        raise
                    logger.exception("Exception in signal handler")

        signal.signal(signalnum, do_handle)
    else:
        handler = __handlers[signalnum]

    # add the callback to the end, but maintain a unique key to remove the callback later
    key = object()
    handler.callbacks[key] = callback
    try:
        yield
    finally:
        del handler.callbacks[key]
        if len(handler.callbacks) == 0 or (len(handler.callbacks) == 1 and None in handler.callbacks):
            # there are no more handlers, or just the old handler
            signal.signal(signalnum, handler.old_handler)
            del __handlers[signalnum]


@contextmanager
def register_async(signalnum, callback):
    if threading.current_thread() is not threading.main_thread():
        raise RuntimeError("Signal handlers can only be manipulated from the main thread")

    if signalnum in __handlers:
        raise RuntimeError("Can't combine regular and event loop signal handling")

    loop = asyncio.get_event_loop()

    if signalnum not in __async_handlers:
        old_handler = signal.getsignal(signalnum)
        if old_handler is None:
            # None means that the previous signal handler was not installed from Python
            # it's not legal to pass None to signal(), so restore the default
            logger.warning("Removing a signal handler that can't be restored")
            if signalnum == signal.SIGINT:
                old_handler = signal.default_int_handler
            else:
                old_handler = signal.SIG_DFL

        handler = __async_handlers[signalnum] = __Handler(old_handler, OrderedDict())

        def do_handle():
            for callback in reversed(__async_handlers[signalnum].callbacks.values()):
                try:
                    callback()
                except Exception:
                    logger.exception("Exception in signal handler")

        loop.add_signal_handler(signalnum, do_handle)
    else:
        handler = __async_handlers[signalnum]

    # add the callback to the end, but maintain a unique key to remove the callback later
    key = object()
    handler.callbacks[key] = callback
    try:
        yield
    finally:
        del handler.callbacks[key]
        if len(handler.callbacks) == 0:
            # there are no more handlers
            loop.remove_signal_handler(signalnum)
            signal.signal(signalnum, handler.old_handler)
            del __async_handlers[signalnum]



from typing import Awaitable, Callable, List

import pytest

import asyncio
import zmq.asyncio
from contextlib import contextmanager, asynccontextmanager

from concurrent_utils.event_loop_thread import EventLoopThread
from hedgehog.protocol import ServerSide
from hedgehog.protocol.zmq.asyncio import DealerRouterSocket
from hedgehog.protocol.messages import Message, ack, analog, digital, imu, io, motor, servo, process, speaker


@pytest.fixture
def start_dummy(zmq_aio_ctx: zmq.asyncio.Context):
    @asynccontextmanager
    async def do_start(server_coro: Callable[[DealerRouterSocket], Awaitable[None]], *args,
                       endpoint: str='inproc://controller', **kwargs):
        with DealerRouterSocket(zmq_aio_ctx, zmq.ROUTER, side=ServerSide) as socket:
            socket.bind(endpoint)

            async def target():
                await server_coro(socket, *args, **kwargs)

                ident, msgs = await socket.recv_msgs()
                _msgs = []  # type: List[Message]
                _msgs.extend(motor.Action(port, motor.POWER, 0) for port in range(0, 4))
                _msgs.extend(servo.Action(port, False, 0) for port in range(0, 4))
                assert msgs == tuple(_msgs)
                await socket.send_msgs(ident, [ack.Acknowledgement()] * 8)

            task = asyncio.ensure_future(target())
            try:
                yield endpoint
                await task
            finally:
                task.cancel()

    return do_start


@pytest.fixture
def start_dummy_sync(start_dummy):
    @contextmanager
    def do_start(server_coro: Callable[[DealerRouterSocket], Awaitable[None]], *args,
                 endpoint: str='inproc://controller', **kwargs):
        with EventLoopThread() as looper, \
                looper.context(start_dummy(server_coro, *args, endpoint=endpoint, **kwargs)) as dummy:
            yield dummy

    return do_start


class Commands(object):
    @staticmethod
    async def concurrent_analog_digital_requests(server, port_a, value_a, port_d, value_d):
        async def handle_a(ident, msg):
            assert msg == analog.Request(port_a)
            await server.send_msg(ident, analog.Reply(port_a, value_a))

        async def handle_d(ident, msg):
            assert msg == digital.Request(port_d)
            await server.send_msg(ident, digital.Reply(port_d, value_d))

        ident, msg = await server.recv_msg()
        if isinstance(msg, analog.Request):
            await handle_a(ident, msg)
            ident, msg = await server.recv_msg()
            await handle_d(ident, msg)
        else:
            await handle_d(ident, msg)
            ident, msg = await server.recv_msg()
            await handle_a(ident, msg)

    @staticmethod
    async def unsupported(server):
        ident, msg = await server.recv_msg()
        await server.send_msg(ident, ack.Acknowledgement(ack.UNSUPPORTED_COMMAND))

    @staticmethod
    async def imu_rate_request(server, x, y, z):
        ident, msg = await server.recv_msg()
        assert msg == imu.RateRequest()
        await server.send_msg(ident, imu.RateReply(x, y, z))

    @staticmethod
    async def imu_acceleration_request(server, x, y, z):
        ident, msg = await server.recv_msg()
        assert msg == imu.AccelerationRequest()
        await server.send_msg(ident, imu.AccelerationReply(x, y, z))

    @staticmethod
    async def imu_pose_request(server, x, y, z):
        ident, msg = await server.recv_msg()
        assert msg == imu.PoseRequest()
        await server.send_msg(ident, imu.PoseReply(x, y, z))

    @staticmethod
    async def io_action_input(server, port, pullup):
        ident, msg = await server.recv_msg()
        assert msg == io.Action(port, io.INPUT_PULLUP if pullup else io.INPUT_FLOATING)
        await server.send_msg(ident, ack.Acknowledgement())

    @staticmethod
    async def io_command_request(server, port, flags):
        ident, msg = await server.recv_msg()
        assert msg == io.CommandRequest(port)
        await server.send_msg(ident, io.CommandReply(port, flags))

    @staticmethod
    async def analog_request(server, port, value):
        ident, msg = await server.recv_msg()
        assert msg == analog.Request(port)
        await server.send_msg(ident, analog.Reply(port, value))

    @staticmethod
    async def digital_request(server, port, value):
        ident, msg = await server.recv_msg()
        assert msg == digital.Request(port)
        await server.send_msg(ident, digital.Reply(port, value))

    @staticmethod
    async def io_action_output(server, port, level):
        ident, msg = await server.recv_msg()
        assert msg == io.Action(port, io.OUTPUT_ON if level else io.OUTPUT_OFF)
        await server.send_msg(ident, ack.Acknowledgement())

    @staticmethod
    async def motor_config_action(server, port, config):
        ident, msg = await server.recv_msg()
        assert msg == motor.ConfigAction(port, config)
        await server.send_msg(ident, ack.Acknowledgement())

    @staticmethod
    async def motor_action(server, port, state, amount):
        ident, msg = await server.recv_msg()
        assert msg == motor.Action(port, state, amount)
        await server.send_msg(ident, ack.Acknowledgement())

    @staticmethod
    async def motor_command_request(server, port, state, amount):
        ident, msg = await server.recv_msg()
        assert msg == motor.CommandRequest(port)
        await server.send_msg(ident, motor.CommandReply(port, motor.DcConfig(), state, amount))

    @staticmethod
    async def motor_state_request(server, port, velocity, position):
        ident, msg = await server.recv_msg()
        assert msg == motor.StateRequest(port)
        await server.send_msg(ident, motor.StateReply(port, velocity, position))

    @staticmethod
    async def motor_set_position_action(server, port, position):
        ident, msg = await server.recv_msg()
        assert msg == motor.SetPositionAction(port, position)
        await server.send_msg(ident, ack.Acknowledgement())

    @staticmethod
    async def servo_action(server, port, active, position):
        ident, msg = await server.recv_msg()
        assert msg == servo.Action(port, active, position)
        await server.send_msg(ident, ack.Acknowledgement())

    @staticmethod
    async def servo_command_request(server, port, active, position):
        ident, msg = await server.recv_msg()
        assert msg == servo.CommandRequest(port)
        await server.send_msg(ident, servo.CommandReply(port, active, position))

    @staticmethod
    async def execute_process_echo_asdf(server, pid):
        ident, msg = await server.recv_msg()
        assert msg == process.ExecuteAction('echo', 'asdf')
        await server.send_msg(ident, process.ExecuteReply(pid))
        await server.send_msg(ident, process.StreamUpdate(pid, process.STDOUT, b'asdf\n'))
        await server.send_msg(ident, process.StreamUpdate(pid, process.STDOUT))
        await server.send_msg(ident, process.StreamUpdate(pid, process.STDERR))
        await server.send_msg(ident, process.ExitUpdate(pid, 0))

    @staticmethod
    async def execute_process_cat(server, pid):
        ident, msg = await server.recv_msg()
        assert msg == process.ExecuteAction('cat')
        await server.send_msg(ident, process.ExecuteReply(pid))

        while True:
            ident, msg = await server.recv_msg()
            chunk = msg.chunk
            assert msg == process.StreamAction(pid, process.STDIN, chunk)
            await server.send_msg(ident, ack.Acknowledgement())

            await server.send_msg(ident, process.StreamUpdate(pid, process.STDOUT, chunk))

            if chunk == b'':
                break

        await server.send_msg(ident, process.StreamUpdate(pid, process.STDERR))
        await server.send_msg(ident, process.ExitUpdate(pid, 0))

    @staticmethod
    async def speaker_action(server, frequency):
        ident, msg = await server.recv_msg()
        assert msg == speaker.Action(frequency)
        await server.send_msg(ident, ack.Acknowledgement())




from contextlib import contextmanager
from pycreate2 import Create2
from .sync_client import HedgehogClient, connect


@contextmanager
def connect_create(port='/dev/ttyUSB0', baud=57600):
    create = Create2(port, baud=baud)
    create.start()
    yield create
    # TODO do proper cleanup instead of waiting for the GC
    # del create does not seem to work, and create.__del__() raises an exception upon actual GC


@contextmanager
def connect_create2(port='/dev/ttyUSB0', baud=115200):
    create = Create2(port, baud=baud)
    create.start()
    yield create
    # TODO do proper cleanup instead of waiting for the GC
